/*******************************************************************************
 * Copyright (C) 2011 Mikołaj Rybiński
 *
 *     This file is part of Tav4SB.
 *
 *     Tav4SB is free software: you can redistribute it and/or modify it under
 *     the terms of the GNU Affero General Public License as published by the
 *     Free Software Foundation, either version 3 of the License, or (at your
 *     option) any later version.
 *
 *     Tav4SB is distributed in the hope that it will be useful, but WITHOUT ANY
 *     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *     FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
 *     more details.
 *
 *     You should have received a copy of the GNU Affero General Public License
 *     along with Tav4SB. If not, see <http://www.gnu.org/licenses/>.
 *
 *     Contributors:
 *         Paweł Banasik - initial implementation
 *         Michał Lula - grid implementation
 *         Mikołaj Rybiński - implementation and maintenance
 ******************************************************************************/
package pl.edu.mimuw.bioputer.tav4sb.workflows.bsh;

import java.math.BigDecimal;
import java.util.LinkedList;
import java.util.List;

import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamWriter;

class MathType {
    String type;
    Object value;
}
//class MathString extends MathType {
//    final String type = STRING_TYPE;
//    String value;
//}
//class MathNumber extends MathType {
//    final String type = NUMBER_TYPE;
//    BigDecimal value;
//}
//class MathSymbol extends MathType {
//    final String type = SYMBOL_TYPE;
//    String value;
//}
//class MathList extends MathType {
//    final String type = LIST_TYPE;
//    List<MathType> value;
//}

class PartiallyParsedListObj {
    MathType head;
    String tail;
}

/**
 * @author trybik
 *
 */
public class LibMathType {

    public static String LIST_TYPE = "List";
    public static String NUMBER_TYPE = "Number";
    public static String SYMBOL_TYPE = "Symbol";
    public static String STRING_TYPE = "String";

    private static String STRING_DELIM = "\"";
    private static String LIST_START = "{";
    private static String LIST_END = "}";
    private static String LIST_EL_SEP = ",";
    private static String ESCAPE_CHAR = "\\";

    private static String mathStringUnescape(String mathStringStr) {
        // string delimiter escaping is optional
        return mathStringStr.replace(ESCAPE_CHAR+LIST_EL_SEP, LIST_EL_SEP)
            .replace(ESCAPE_CHAR+LIST_START, LIST_START)
            .replace(ESCAPE_CHAR+LIST_END, LIST_END)
            .replace(ESCAPE_CHAR+STRING_DELIM, STRING_DELIM);
    }

    /**
     * Same as the String method indexOf, but returns the length of the string
     * param instead of -1, if the separator wasn't found.
     */
    private static int indexOf(String str, String sep) {
        int ret = str.indexOf(sep);
        return (ret < 0) ? str.length() : ret;
    }

    /**
     * Get the first index of a non-escaped sep in str. If not found returns the
     * length of the string which is not a valid index but is a valid arg for
     * the substring method.
     */
    private static int unescapedIndexOf(String str, String sep) {
        // we're interested in non-escaped separators
        if (str.startsWith(sep))
            return 0;
        int i = -1; // sep start index
        while (i < str.length()-1) {
            i = str.indexOf(sep, i+1);
            if (i < 0)
                i = str.length();
            else if (!str.startsWith(ESCAPE_CHAR, i-1))
                break;
        }
        return i;
    }

    private static PartiallyParsedListObj partiallyParsedList(String tail) {
        MathType head = new MathType();
        if (tail.startsWith(LIST_START)) {
            // extract inner list element
            tail = tail.substring(1); // cut the opening bracket
            PartiallyParsedListObj ppInnerList;
            List<MathType> innerList = new LinkedList<MathType>();
            while (!tail.startsWith(LIST_END)) { // stop at the closing bracket
                ppInnerList = partiallyParsedList(tail);
                innerList.add(ppInnerList.head);
                tail = ppInnerList.tail;
                if (tail.length() == 0) {
                    throw new ExceptionMessage("missing closing list bracket '}'");
                }
            }
            tail = tail.substring(1).trim(); // cut the closing bracket
            head.type = LIST_TYPE;
            head.value = innerList;
        } else {
            // extract basic type element
            int splitIndex;
            if (tail.startsWith(STRING_DELIM)) {
                // extract element by jumping to the end of the string delimiter
                splitIndex = Math.min(unescapedIndexOf(tail.substring(1),
                        STRING_DELIM) + 2, tail.length());
            } else {
                // extract element by splitting tail with a non-escaped comma or
                // a closing bracket
                splitIndex = Math.min(indexOf(tail, LIST_EL_SEP),
                        indexOf(tail, LIST_END));
            }
            head = mathtype(tail.substring(0, splitIndex));
            tail = tail.substring(splitIndex).trim();
        }
        // if necessary, cut the list element separator
        if (tail.length() > 0 && !tail.startsWith(LIST_END)) {
            if (tail.startsWith(LIST_EL_SEP)) {
                tail = tail.substring(1).trim();
            } else {
                throw new ExceptionMessage("missing list element separator ',' near '" + tail.substring(0, Math.min(10, tail.length())) + "...'");
            }
        }

        PartiallyParsedListObj ret = new PartiallyParsedListObj();
        ret.head = head;
        ret.tail = tail;
        return ret;
    }

    public static MathType mathtype(String mathtypestr) {
        String in = mathtypestr.trim();
        String type = null;
        Object value = null;
        if (in.startsWith(STRING_DELIM)) {
            if (!(in.length()>1 && in.endsWith(STRING_DELIM))) {
                throw new ExceptionMessage("missing closing string quote '\"'");
            }
            type = STRING_TYPE;
            value = mathStringUnescape(in.substring(1, in.length()-1));
        } else if (in.startsWith(LIST_START)) {
            if (!in.endsWith(LIST_END)) {
                throw new ExceptionMessage("missing closing list bracket '}'");
            }
            type = LIST_TYPE;
            List<MathType> list = new LinkedList<MathType>();
            // manual parsing cuz of nested lists
            PartiallyParsedListObj partialyParsedList;
            in = in.substring(1, in.length()-1).trim();
            while (in.length()>0) {
                partialyParsedList = partiallyParsedList(in);
                in = partialyParsedList.tail;
                list.add(partialyParsedList.head);
            }
            value = list;
        } else if (in.contains(".")) {
            type = NUMBER_TYPE;
            try {
                value = new BigDecimal(in);
            } catch (NumberFormatException e) {
                throw new ExceptionMessage("number '" + in + "' format exception");
            }
        } else {
            type = SYMBOL_TYPE;
            value = in;
        }

        MathType ret = new MathType();
        ret.type = type;
        ret.value = value;
        return ret;
    }



    public static String XSI_NS_PREFIX = "xsi";
    public static String XSI_NS_URI = "http://www.w3.org/2001/XMLSchema-instance";
    public static String MATH_NS_PREFIX = "math";
    static String MATH_NS_URI = "http://bioputer.mimuw.edu.pl/tav4sb/services/xsd/mathematica";
    static String EL_TAGNAME = "element";

    public static void writeMathTypeElement(XMLStreamWriter xw, MathType mathtype, String elname) throws XMLStreamException {
        xw.writeStartElement(elname);
        xw.writeAttribute(XSI_NS_PREFIX, XSI_NS_URI, "type", MATH_NS_PREFIX+":"+mathtype.type);
        if (mathtype.type.equals(LIST_TYPE)) {
            for (MathType elmathtype: (List<MathType>) mathtype.value) {
                writeMathTypeElement(xw, elmathtype, EL_TAGNAME);
            }
        } else {
            xw.writeStartElement(mathtype.type.toLowerCase());
            xw.writeCharacters(mathtype.value.toString());
            xw.writeEndElement();
        }
        xw.writeEndElement();
    }

}
