/*******************************************************************************
 * Copyright (C) 2011 Mikołaj Rybiński
 * 
 *     This file is part of Tav4SB.
 * 
 *     Tav4SB is free software: you can redistribute it and/or modify it under the
 *     terms of the GNU Affero General Public License as published by the Free
 *     Software Foundation, either version 3 of the License, or (at your option)
 *     any later version.
 * 
 *     Tav4SB is distributed in the hope that it will be useful, but WITHOUT ANY
 *     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *     FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
 *     more details.
 * 
 *     You should have received a copy of the GNU Affero General Public License
 *     along with Tav4SB. If not, see <http://www.gnu.org/licenses/>.
 * 
 *     Contributors:
 *         Paweł Banasik - initial implementation
 *         Michał Lula - grid implementation
 *         Mikołaj Rybiński - implementation and maintenance
 ******************************************************************************/
package pl.edu.mimuw.bioputer.tav4sb.consumer.task;

import java.util.HashMap;
import java.util.Map;
import javax.jms.MapMessage;
import javax.jms.Message;
import org.apache.log4j.Logger;
import pl.edu.mimuw.bioputer.tav4sb.consumer.Consumer;
import pl.edu.mimuw.bioputer.tav4sb.utils.Throwable;

/**
 *
 * @author michal
 */
public class TaskConsumer extends Consumer {

    private final Map<String, WorkerFactory> factories = new HashMap<String, WorkerFactory>();
    private static final Logger logger = Logger.getLogger(TaskConsumer.class);

    public TaskConsumer(String brokerURL, String queueName) {
        super(brokerURL, queueName);
    }

    public void onMessage(Message message) {
        try {
            Map<String, String> reply = new HashMap<String, String>();
            if (message instanceof MapMessage) {
                MapMessage map = (MapMessage) message;
                if (map.itemExists("TYPE") && map.itemExists("TEXT")) {
                    logger.info("Message["+message.getJMSType()+"]; id=" + message.getJMSMessageID() + "; type=" + map.getString("TYPE"));
                    logger.debug("message " + message.getJMSMessageID() + " text:\n" + map.getString("TEXT"));
                    String type = map.getString("TYPE");
                    String text = map.getString("TEXT");
                    try {
                        if (getFactories().containsKey(type)) {
                                Worker worker = getWorker(type);
                                reply.put("COMPUTATION_START", String.valueOf(System.currentTimeMillis()));
                                String result = worker.execute(text);
                                reply.put("COMPUTATION_END", String.valueOf(System.currentTimeMillis()));
                                logger.debug(result);
                                reply.put("RESULT", result);
                        } else {
                            throw new UnregisteredTaskTypeException(type);
                        }
                    } catch (Exception e) {
                        reply.put("EXCEPTION", Throwable.toString(e));
                    }
                    if (map.itemExists("ID")) {
                        reply.put("ID", map.getString("ID"));
                    }
                }
            }
            reply(message, reply);
            logger.info("message " + message.getJMSMessageID() +
                    " processing completed " + (reply.containsKey("EXCEPTION") ?
                        "with EXCEPTION\n" + reply.get("EXCEPTION") :
                        "SUCCESSFULLY"));
        } catch (Exception e) {
            logger.error("Exception occured while processing message. Check if worker is up and working.", e);
        }
    }

    private Worker getWorker(String type) {
        return getFactories().get(type).getWorker();
    }

    public void registerWorkerFactory(String key, WorkerFactory factory) {
        logger.info("Registering factory of type: " + key);
        getFactories().put(key, factory);
    }

    private Map<String, WorkerFactory> getFactories() {
        return this.factories;
    }
}
