/*******************************************************************************
 * Copyright (C) 2011 Mikołaj Rybiński
 *
 *     This file is part of Tav4SB.
 *
 *     Tav4SB is free software: you can redistribute it and/or modify it under the
 *     terms of the GNU Affero General Public License as published by the Free
 *     Software Foundation, either version 3 of the License, or (at your option)
 *     any later version.
 *
 *     Tav4SB is distributed in the hope that it will be useful, but WITHOUT ANY
 *     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *     FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
 *     more details.
 *
 *     You should have received a copy of the GNU Affero General Public License
 *     along with Tav4SB. If not, see <http://www.gnu.org/licenses/>.
 *
 *     Contributors:
 *         Paweł Banasik - initial implementation
 *         Michał Lula - grid implementation
 *         Mikołaj Rybiński - implementation and maintenance
 ******************************************************************************/
package pl.edu.mimuw.bioputer.tav4sb.logic.mathematica.proc.plot;

import com.wolfram.jlink.Expr;
import com.wolfram.jlink.MathLinkException;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.log4j.Logger;
import pl.edu.mimuw.bioputer.tav4sb.logic.mathematica.engine.MEngine;
import pl.edu.mimuw.bioputer.tav4sb.logic.mathematica.engine.MEngineException;
import pl.edu.mimuw.bioputer.tav4sb.logic.mathematica.engine.MEngineUtils;
import pl.edu.mimuw.bioputer.tav4sb.logic.mathematica.proc.FunctionProc;
import pl.edu.mimuw.bioputer.tav4sb.logic.mathematica.utils.XsdMathTypeUtils;
import pl.edu.mimuw.bioputer.tav4sb.services.xsd.mathematica.MathType;

/**
 *
 * @author michal
 */
public class Plot extends FunctionProc {

    private static final Logger logger = Logger.getLogger(Plot.class);
    private final List<AbstractPlot> plots = new LinkedList<AbstractPlot>();
    private final Map<String, MathType> options = new HashMap<String, MathType>();
    private final int dimension;
    private byte[] data = null;
    public static final String MATH_PLOT_2D = "MathPlot2D";
    public static final String MATH_PLOT_3D = "MathPlot3D";
    private static final String[] sourceFiles = {"Utilz.m", "mathPlot.m"};

    public Plot(int dimension) {
        this.dimension = dimension;
    }

    public Plot(int dimension, Map<String, MathType> options) {
        this(dimension);
        this.options.putAll(options);
    }

    public void addPlots(List<AbstractPlot> plots) {
        for (AbstractPlot plot : plots) {
            addPlot(plot);
        }
    }

    public int getDimension() {
        return this.dimension;
    }

    private Map<String, MathType> getOptions() {
        return options;
    }

    public List<AbstractPlot> getPlots() {
        return this.plots;
    }

    public void addPlot(AbstractPlot plot) {
        if (plot.getDimension() == getDimension()) {
            getPlots().add(plot);
        } else {
            String errMsg = "Added plot have an incompatible number of dimensions.";
            logger.error(errMsg);
            throw new IllegalArgumentException(errMsg);
        }
    }

    protected Expr plotsToExpr() {
        List<Expr> list = new LinkedList<Expr>();
        for (AbstractPlot plot : getPlots()) {
            list.add(plot.toExpr());
        }
        return MEngineUtils.list(list);
    }

    @Override
    protected String getFunctionName() {
        switch (getDimension()) {
            case 2:
                return MATH_PLOT_2D;
            case 3:
                return MATH_PLOT_3D;
            default:
                throw new IllegalArgumentException("Unsuported dimensions number: " + Integer.toString(getDimension()) + ".");
        }
    }

    @Override
    protected Expr[] getArgs() {
        //new Expr[]{plotsToExpr(), argsLabelsToExpr()}
        Expr[] args = new Expr[1+getOptions().size()];
        args[0] = plotsToExpr();
        // TODO: instead of hand-made mathematica value types, consider passing
        //       any string and then setting it as rhs of rule with
        //       'ToExpression' Mathematica function; consider saftiness, i.e.
        //       kernel evaluation of Mathematica code from string should be
        //       done in sandbox environment
        int i = 1;
        for (Entry<String, MathType> opt : getOptions().entrySet()) {
            args[i] = MEngineUtils.rule(MEngineUtils.symbol(opt.getKey()),
                XsdMathTypeUtils.toExpr(opt.getValue()));
            i++;
        }
        return args;
    }

    /**
     * @return the data
     */
    public byte[] getData() {
        return data;
    }

    public void setData(byte[] b) {
        this.data = b;
    }

    /**
     * save plot content on hard drive if plot content is not empty
     * @param file
     * @throws java.io.IOException
     */
    public void save(File file) throws IOException {
        if (getData() != null) {
            FileOutputStream fos = new FileOutputStream(file);
            fos.write(getData());
            fos.close();
        }
    }

    @Override
    protected String[] getSourceFiles() {
        return sourceFiles;
    }
    @Override
    protected byte[] evaluateExpr(MEngine engine, Expr expr) throws MathLinkException {
        return engine.plot(expr);
    }
    public void draw() throws MathLinkException, MEngineException {
        setData((byte[]) evaluate());
    }

    @Override
    public String toString() {
        if (getData() != null) {
            return "size(" + getData().length + ") data: " + getData();
        } else {
            return "empty plot";
        }
    }
}
