/*******************************************************************************
 * Copyright (C) 2011 Mikołaj Rybiński
 * 
 *     This file is part of Tav4SB.
 * 
 *     Tav4SB is free software: you can redistribute it and/or modify it under the
 *     terms of the GNU Affero General Public License as published by the Free
 *     Software Foundation, either version 3 of the License, or (at your option)
 *     any later version.
 * 
 *     Tav4SB is distributed in the hope that it will be useful, but WITHOUT ANY
 *     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *     FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
 *     more details.
 * 
 *     You should have received a copy of the GNU Affero General Public License
 *     along with Tav4SB. If not, see <http://www.gnu.org/licenses/>.
 * 
 *     Contributors:
 *         Paweł Banasik - initial implementation
 *         Michał Lula - grid implementation
 *         Mikołaj Rybiński - implementation and maintenance
 ******************************************************************************/
package pl.edu.mimuw.bioputer.tav4sb.producer;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.jms.IllegalStateException;
import javax.jms.JMSException;

import org.apache.log4j.Logger;

import pl.edu.mimuw.bioputer.tav4sb.producer.utils.ProducerProperties;

/**
 *
 * @author michal
 */
public abstract class ServiceMessage {

    private static final Logger logger = Logger.getLogger(ServiceMessage.class);

    private static final ProducerProperties props = new ProducerProperties();

    protected abstract List<String> getTextList();
    protected abstract void addResult(String result);
    protected abstract String getQueueName();
    protected abstract String getRecorderQueueName();
    protected abstract String getReceiverQueueName();
    protected abstract String getMessageType();

    public abstract String getBrokerUrl();

    private void doSendTasks(Producer receiver, String id) throws JMSException {
        logger.debug("sending tasks to receiver: brokerUrl="+receiver.getBrokerUrl()+", queueName="+receiver.getQueueName()+".");
        for (String text : getTextList()) {
            receiver.send(getMessageType(), text, id);
            logger.debug("task(id="+ id +") message("+ getMessageType() +") sent to receiver:\n"+ text);
        }
    }

    private void sendTasks() throws JMSException {
        Producer receiver = new Producer(getBrokerUrl(), getQueueName());
        try {
            doSendTasks(receiver, Producer.ANONYMOUS_TASK_ID);
            // get results synchornously
            for (int i = 0; i < getTextList().size(); i++) {
                addResult(receiver.receive());
                logger.debug("received result #"+Integer.toString(i+1)+".");
            }
        } finally {
            receiver.close();
        }
    }

    private void sendTasks(String id) throws JMSException {
        Producer receiver = new Producer(getBrokerUrl(), getQueueName(), getReceiverQueueName());
        try {
            doSendTasks(receiver, id);
        } finally {
            receiver.close();
        }
    }

    public void process() throws JMSException {
        logger.debug("begin");
        sendTasks();
        logger.debug("end");
    }

    private String registerTasks(String email) throws JMSException {
        String id = null;
        Producer recorder = new Producer(getBrokerUrl(), getRecorderQueueName());
        logger.debug("recorder created: brokerUrl="+recorder.getBrokerUrl()+", queueName="+recorder.getQueueName()+".");
        try {
            Map<String, String> map = new HashMap<String, String>();
            map.put("TYPE", getMessageType());
            map.put("QUANTITY", String.valueOf(getTextList().size()));
            map.put("EMAIL", email);
            recorder.send(map);
            logger.debug("Sent tasks recording request");
            id = recorder.receive(props.getAsyncRegisterTimeout());
            logger.debug("task(id="+ id +") registered in recorder");
        } catch (Exception e) {
            StringWriter emw = new StringWriter();
            emw.write("Couldn't register asynchronous task: type="+getMessageType());
            emw.write("; email="+email);
            emw.write(". Internal exception:\n");
            emw.write(e.getMessage()+"\n");
            e.printStackTrace(new PrintWriter(emw));
            String errMsg = emw.toString();
            logger.error(errMsg);
            throw new IllegalStateException(errMsg);
        } finally {
            recorder.close();
        }
        return id;
    }

    public void asyncProcess(String email) throws JMSException {
        logger.debug("begin");
        sendTasks(registerTasks(email));
        logger.debug("end");
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        for (String s : getTextList()) {
            sb.append("message:\n");
            sb.append(s);
            sb.append("\n");
        }
        return sb.toString();
    }
}
