/*******************************************************************************
 * Copyright (C) 2011 Mikołaj Rybiński
 * 
 *     This file is part of Tav4SB.
 * 
 *     Tav4SB is free software: you can redistribute it and/or modify it under the
 *     terms of the GNU Affero General Public License as published by the Free
 *     Software Foundation, either version 3 of the License, or (at your option)
 *     any later version.
 * 
 *     Tav4SB is distributed in the hope that it will be useful, but WITHOUT ANY
 *     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *     FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
 *     more details.
 * 
 *     You should have received a copy of the GNU Affero General Public License
 *     along with Tav4SB. If not, see <http://www.gnu.org/licenses/>.
 * 
 *     Contributors:
 *         Paweł Banasik - initial implementation
 *         Michał Lula - grid implementation
 *         Mikołaj Rybiński - implementation and maintenance
 ******************************************************************************/
package pl.edu.mimuw.bioputer.tav4sb.register.comp;

import java.io.File;
import java.io.IOException;
import java.text.ParseException;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;
import javax.jms.MapMessage;
import javax.jms.Message;
import javax.xml.bind.JAXBException;

import org.apache.log4j.Logger;

import pl.edu.mimuw.bioputer.tav4sb.register.Computation;
import pl.edu.mimuw.bioputer.tav4sb.utils.file.RemovableFilesHolder;
import pl.edu.mimuw.bioputer.tav4sb.utils.file.SaveableString;
import pl.edu.mimuw.bioputer.tav4sb.utils.xml.XmlTranslator;

/**
 *
 * @author michal
 */
public abstract class AbstractComputation<T> implements Computation {

    private static final Logger logger = Logger
            .getLogger(AbstractComputation.class);

    private final String email;
    private final String id;
    private final int quantity;
    private int currentQuantity = 0;

    private final List<Message> answerList = new LinkedList<Message>();
    private final List<String> errorList = new LinkedList<String>();

    private final XmlTranslator<T, T> translator;
    private RemovableFilesHolder resultHolder;

    private final Object updateTimingLock = new Object();
    private TimeInterval globalTiming;
    private long longest = 0;
    private long pooled = 0;
    private List<TimeInterval> timingsList = new LinkedList<TimeInterval>();

    abstract protected String getContext();

    public AbstractComputation(String id, int quantity, String email) {
        this.translator = new XmlTranslator<T, T>(getContext());
        this.id = id;
        this.quantity = quantity;
        this.email = email;
    }

    private void updateTimings(String start, String end) throws ParseException {
        Date startDate = new Date(Long.parseLong(start));
        Date endDate = new Date(Long.parseLong(end));
        TimeInterval timing = new TimeInterval(startDate, endDate);
        synchronized (updateTimingLock) {
            updateGlobalTiming(startDate, endDate);
            setLongest(timing.getDuration());
            addDuration(timing.getDuration());
            getTimingsList().add(timing);
        }
    }

    private void updateGlobalTiming(Date start, Date end) {
        if (globalTiming == null) {
            globalTiming = new TimeInterval(start, end);
        } else {
            globalTiming.extend(start, end);
        }
    }

    private TimeInterval getGlobalTiming() {
        return this.globalTiming;
    }

    private List<TimeInterval> getTimingsList() {
        return this.timingsList;
    }

    private void setLongest(long longest) {
        if (getLongest() < longest) {
            this.longest = longest;
        }
    }

    private long getLongest() {
        return this.longest;
    }

    private void addDuration(long duration) {
        this.pooled += duration;
    }

    private long getPooled() {
        return this.pooled;
    }

    public void addAnswer(Message answer) throws Exception {
        incCurrentQuantity();
        getAnswers().add(answer);
        if (answer instanceof MapMessage) {
            MapMessage message = (MapMessage) answer;
            if (message.itemExists("EXCEPTION")) {
                addError(message.getString("EXCEPTION"));
            } else {
                if (message.itemExists("COMPUTATION_START")
                        && message.itemExists("COMPUTATION_END")) {
                    updateTimings(message.getString("COMPUTATION_START"),
                            message.getString("COMPUTATION_END"));
                }
                if (message.itemExists("RESULT")) {
                    String result = message.getString("RESULT");
                    setResultObject(getTranslator().stringToObject(result));
                } else {
                    addError("No result");
                }

            }
        } else {
            addError("Not supported type of message");
        }
    }

    public List<Message> getAnswers() {
        return this.answerList;
    }

    public abstract void setResultObject(T answer);

    public List<String> getErrors() {
        return this.errorList;
    }

    public void addError(String message) {
        getErrors().add(message);
    }

    public boolean isComplete() {
        return (getQuantity() == getCurrentQuantity());
    }

    public String getResult() {
        try {
            return getTranslator().objectToString(getResultObject());
        } catch (JAXBException ex) {
            String errMsg = "Couldn't translate computation results object to String.";
            logger.error(errMsg, ex);
            throw new RuntimeException(errMsg, ex);
        }
    }

    public abstract T getResultObject();

    private void setResultHolder(RemovableFilesHolder resultHolder) {
        this.resultHolder = resultHolder;
    }

    private RemovableFilesHolder getResultHolder() {
        return resultHolder;
    }

    private File createTempResultFile() {
        try {
            File resultFile = File.createTempFile("tav4sb_"
                    + getComputationType() + "_result_", ".xml");
            logger.debug("Created temporary computation result file: "
                    + resultFile.getAbsolutePath());
            setResultHolder(new RemovableFilesHolder(resultFile));
            new SaveableString(getResult()).save(resultFile);
            return resultFile;
        } catch (IOException ex) {
            String errMsg = "Couldn't create or write to temporary computation result file.";
            logger.error(errMsg, ex);
            throw new RuntimeException(errMsg, ex);
        }
    }

    public File getResultFile() {
        return createTempResultFile();
    }

    public String getStatistic() {
        StringBuffer sb = new StringBuffer();
        sb.append("Computation type: " + getComputationType() + "\n");
        sb.append("Actual computations quantity: " + getQuantity() + "\n");
        sb.append("Nr of completed computations: " + getCurrentQuantity()
                + "\n");
        sb.append("Started at: " + getGlobalTiming().getStart() + "\n");
        sb.append("Ended at: " + getGlobalTiming().getEnd() + "\n");
        sb.append("Real-time duration: " + getGlobalTiming().getDuration()
                + " ms\n");
        sb.append("Longest computation duration: " + getLongest() + " ms\n");
        sb.append("Pooled duration: " + getPooled() + " ms\n");
        sb.append("Timing details:\n");
        List<TimeInterval> list = getTimingsList();
        for (int i = 0; i < list.size(); i++) {
            TimeInterval seDate = list.get(i);
            sb.append("- [Comp. #" + String.valueOf(i) + "] start: "
                    + seDate.getStart() + "; end: " + seDate.getEnd()
                    + "; duration: " + seDate.getDuration() + " ms;\n");

        }
        List<String> errs = getErrors();
        sb.append("Errors count: " + errs.size()
                + (errs.size() > 0 ? "; details:" : "") + "\n");
        for (int i = 0; i < errs.size(); i++) {
            sb.append("- [Error #" + String.valueOf(i) + "] " + errs.get(i)
                    + "\n");
        }
        return sb.toString();
    }

    /**
     * @return the translator
     */
    public XmlTranslator<T, T> getTranslator() {
        return translator;
    }

    public String getComputationId() {
        return this.id;
    }

    public String getEmail() {
        return this.email;
    }

    public int getQuantity() {
        return this.quantity;
    }

    public int getCurrentQuantity() {
        return this.currentQuantity;
    }

    protected void incCurrentQuantity() {
        this.currentQuantity += 1;
    }
}
