/*******************************************************************************
 * Copyright (C) 2011 Mikołaj Rybiński
 * 
 *     This file is part of Tav4SB.
 * 
 *     Tav4SB is free software: you can redistribute it and/or modify it under the
 *     terms of the GNU Affero General Public License as published by the Free
 *     Software Foundation, either version 3 of the License, or (at your option)
 *     any later version.
 * 
 *     Tav4SB is distributed in the hope that it will be useful, but WITHOUT ANY
 *     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *     FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
 *     more details.
 * 
 *     You should have received a copy of the GNU Affero General Public License
 *     along with Tav4SB. If not, see <http://www.gnu.org/licenses/>.
 * 
 *     Contributors:
 *         Paweł Banasik - initial implementation
 *         Michał Lula - grid implementation
 *         Mikołaj Rybiński - implementation and maintenance
 ******************************************************************************/
package pl.edu.mimuw.bioputer.tav4sb.utils.cli;

import java.io.IOException;
import java.io.InputStream;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import org.apache.cxf.helpers.IOUtils;
import org.apache.log4j.Logger;

/**
 *
 * @author trybik
 */
public class SysCmd {

    private static final Logger logger = Logger.getLogger(SysCmd.class);
    public static final String ERROR_KEY = "err";
    public static final String OUT_KEY = "out";
    private final String cmd;
    private String out;

    /**
     * Content of the standard error stream of the last run of command.
     * Null if command wasn't executed.
     *
     * @return String or null.
     */
    public String getErr() {
        return err;
    }

    private void setErr(String err) {
        this.err = err;
    }

    /**
     * Content of the standard output stream of the last run of command.
     * Null if command wasn't executed.
     *
     * @return String or null.
     */
    public String getOut() {
        return out;
    }

    private void setOut(String out) {
        this.out = out;
    }
    private String err;

    public SysCmd(String cmd) {
        this.cmd = cmd;
    }

    public String getCmd() {
        return this.cmd;
    }

    private static String readInputStream(InputStream in, Boolean close) throws IOException {
        try {
            return IOUtils.toString(in);
        } finally {
            if (close) {
                in.close();
            }
        }
    }
    private static String readInputStream(InputStream in) throws IOException {
        return readInputStream(in, true);
    }

    /**
     * Execute system command. Note: it's up to user of this function to
     * interpret standard out and error streams.
     *
     * @param args String array with command and its options.
     * @return String -> String map with two entries: SysCmd.OUT_KEY with the
     *         standard output generated by the executed command and
     *         SysCmd.ERROR_KEY with the error stream content.
     * @throws java.io.IOException when I/O error occured
     * @throws java.lang.InterruptedException when thread executing command was
     * 										  interrupted before finishing
     * @throws java.lang.RuntimeException when there is output on error stream
     *                                    and nothing on output stream
     */
    public static Map<String,String> execute(String[] args) throws IOException, InterruptedException {
        logger.debug("Runtime exec: " + Arrays.toString(args));
        Process p = Runtime.getRuntime().exec(args);
        String out = readInputStream(p.getInputStream());
        logger.debug("result:\n" + out.toString());
        String err = readInputStream(p.getErrorStream());
        if (!err.isEmpty()) {
            logger.error("errors:\n" + err.toString());
        }
        Map<String,String> ret = new HashMap<String, String>(2);
        ret.put(OUT_KEY, out);
        ret.put(ERROR_KEY, err);
        p.waitFor();
        logger.debug("Runtime exec process ended");
        return ret;
    }

    /**
     * Run SysCmd.getCmd().
     *
     * @param args Commmand arguments as in 'main' method, without the command
     *             name.
     * @return Standard out results.
     * @throws IOException
     * @throws InterruptedException
     */
    public String run(String[] args) throws IOException, InterruptedException {
        String[] strArgs = new String[args.length + 1];
        strArgs[0] = getCmd();
        for (int i = 1; i < strArgs.length; i++) {
            strArgs[i] = args[i - 1];
        }
        Map<String, String> res = execute(strArgs);
        setOut(res.get(OUT_KEY));
        setErr(res.get(ERROR_KEY));
        return getOut();
    }
}
