/*******************************************************************************
 * Copyright (C) 2011 Mikołaj Rybiński
 * 
 *     This file is part of Tav4SB.
 * 
 *     Tav4SB is free software: you can redistribute it and/or modify it under the
 *     terms of the GNU Affero General Public License as published by the Free
 *     Software Foundation, either version 3 of the License, or (at your option)
 *     any later version.
 * 
 *     Tav4SB is distributed in the hope that it will be useful, but WITHOUT ANY
 *     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *     FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
 *     more details.
 * 
 *     You should have received a copy of the GNU Affero General Public License
 *     along with Tav4SB. If not, see <http://www.gnu.org/licenses/>.
 * 
 *     Contributors:
 *         Paweł Banasik - initial implementation
 *         Michał Lula - grid implementation
 *         Mikołaj Rybiński - implementation and maintenance
 ******************************************************************************/
package pl.edu.mimuw.bioputer.tav4sb.utils.props;

import java.io.IOException;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.List;
import org.apache.log4j.Logger;

/**
 * Wrapper for the PropertiesManagerFactory and PropertiesManager. Turns
 * exceptions into IllegalStateException runtime exceptions.
 *
 * @author trybik
 */
public abstract class PropertiesWrapper {

    protected static final Logger logger = Logger.getLogger(PropertiesWrapper.class);

    /**
     * Override this method in subclasses to set the location independent path
     * to the properties file. In most cases you just want to use static
     * variable, e.g.
     *  private static final path = "consumer.properties"
     *  @Override
     *  private String getPath() {
     *      return path;
     *  }
     *
     * @return Path to the properties file.
     */
    abstract protected String getPath();

    private PropertiesManager _getPropertiesManager() throws IOException {
        return PropertiesManagerFactory.createProperitesManager(getPath(), this.getClass().getClassLoader());
    }

    private PropertiesManager getPropertiesManager() {
        try {
            return _getPropertiesManager();
        } catch (IOException ex) {
            String path = getPath();
            logger.error("Cannot load properties file from \"" + path + "\" path.", ex);
            throw new IllegalStateException("Cannot load properties file from \"" + path + "\" path.");

        }
    }

    private String _getProperty(String propertyName) throws PropertyNotFoundException {
        return getPropertiesManager().getProperty(propertyName);
    }

    protected String getProperty(String propertyName) {
        try {
            return _getProperty(propertyName);
        } catch (PropertyNotFoundException ex) {
            logger.error("Cannot find property \"" + propertyName + "\".", ex);
            throw new IllegalStateException("Cannot find property \"" + propertyName + "\".");
        }
    }

    protected String[] getPropertyArray(String key) {
        logger.debug("key = " + key);
        List<String> values = new ArrayList<String>();
        for (int i = 0;; i++) {
            String value = null;
            try {
                value = _getProperty(key+i);
            } catch (PropertyNotFoundException ex) {
                if (i == 0) {
                    logger.error("Cannot find property \"" + key+i + "\".", ex);
                    throw new IllegalStateException("Cannot find property \"" + key+i + "\".");
                }
                break;
            }
            values.add(value);
        }
        return values.toArray(new String[0]);
    }

    protected int getIntProperty(String s) {
        return Integer.parseInt(getProperty(s));
    }

    protected BigDecimal getDecimalProperty(String key) {
        return new BigDecimal(Double.parseDouble(getProperty(key)));
    }
}
